@@ -1,52 +1,88 @@
-#!/usr/bin/env python2
-# vim: set et sw=4:
-"""agi
-
-This module contains functions and classes to implment AGI scripts in python.
+#! /usr/bin/env python
+# -*- coding: utf-8 -*-
+# vim: set et sw=4 fenc=utf-8: 
+"""
+.. module:: agi
+   :synopsis: This module contains functions and classes to implment AGI scripts in python. 
+   
 pyvr
 
 {'agi_callerid' : 'mars.putland.int',
  'agi_channel'  : 'IAX[kputland@kputland]/119',
  'agi_context'  : 'default',
- 'agi_dnid'     : '666',
+ 'agi_dnid'     : '1000',
  'agi_enhanced' : '0.0',
- 'agi_extension': '666',
+ 'agi_extension': '1000',
  'agi_language' : 'en',
  'agi_priority' : '1',
  'agi_rdnis'    : '',
  'agi_request'  : 'pyst',
  'agi_type'     : 'IAX'}
 
+Specification
+-------------
 """
 
-import sys, pprint, re
+import sys
+import pprint
+import re
 import signal
+from six import PY3
 
-DEFAULT_TIMEOUT = 2000 # 2sec timeout used as default for functions that take timeouts
-DEFAULT_RECORD  = 20000 # 20sec record time
+DEFAULT_TIMEOUT = 2000  # 2sec timeout used as default for functions that take timeouts
+DEFAULT_RECORD = 20000  # 20sec record time
 
 re_code = re.compile(r'(^\d*)\s*(.*)')
 re_kv = re.compile(r'(?P<key>\w+)=(?P<value>[^\s]+)\s*(?:\((?P<data>.*)\))*')
 
-class AGIException(Exception): pass
-class AGIError(AGIException): pass
 
-class AGIUnknownError(AGIError): pass
+class AGIException(Exception):
+    pass
 
-class AGIAppError(AGIError): pass
+
+class AGIError(AGIException):
+    pass
+
+
+class AGIUnknownError(AGIError):
+    pass
+
+
+class AGIAppError(AGIError):
+    pass
 
 # there are several different types of hangups we can detect
 # they all are derrived from AGIHangup
-class AGIHangup(AGIAppError): pass
-class AGISIGHUPHangup(AGIHangup): pass
-class AGISIGPIPEHangup(AGIHangup): pass
-class AGIResultHangup(AGIHangup): pass
 
-class AGIDBError(AGIAppError): pass
 
-class AGIUsageError(AGIError): pass
-class AGIInvalidCommand(AGIError): pass
+class AGIHangup(AGIAppError):
+    pass
 
+
+class AGISIGHUPHangup(AGIHangup):
+    pass
+
+
+class AGISIGPIPEHangup(AGIHangup):
+    pass
+
+
+class AGIResultHangup(AGIHangup):
+    pass
+
+
+class AGIDBError(AGIAppError):
+    pass
+
+
+class AGIUsageError(AGIError):
+    pass
+
+
+class AGIInvalidCommand(AGIError):
+    pass
+
+        
 class AGI:
     """
     This class encapsulates communication between Asterisk an a python script.
@@ -54,10 +90,10 @@ class AGI:
     Asterisk.
     """
 
-    def __init__(self,stdin=sys.stdin,stdout=sys.stdout,stderr=sys.stderr):
-        self.stdin=stdin
-        self.stdout=stdout
-        self.stderr=stderr
+    def __init__(self, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
+        self.stdin = stdin
+        self.stdout = stdout
+        self.stderr = stderr
         self._got_sighup = False
         signal.signal(signal.SIGHUP, self._handle_sighup)  # handle SIGHUP
         self.stderr.write('ARGS: ')
@@ -69,13 +105,15 @@ class AGI:
     def _get_agi_env(self):
         while 1:
             line = self.stdin.readline().strip()
+            if PY3:
+                if type(line) is bytes: line = line.decode('utf8')
             self.stderr.write('ENV LINE: ')
             self.stderr.write(line)
             self.stderr.write('\n')
             if line == '':
                 #blank line signals end
                 break
-            key,data = line.split(':')[0], ':'.join(line.split(':')[1:])
+            key, data = line.split(':')[0], ':'.join(line.split(':')[1:])
             key = key.strip()
             data = data.strip()
             if key != '':
@@ -85,7 +123,15 @@ class AGI:
         self.stderr.write('\n')
 
     def _quote(self, string):
-        return ''.join(['"', str(string), '"'])
+        """ provides double quotes to string, converts int/bool to string """
+        if isinstance(string, int):
+          string = str(string)
+        if isinstance(string, float):
+          string = str(string)
+        if PY3:
+            return ''.join(['"', string, '"'])
+        else:
+            return ''.join(['"', string.encode('utf8', 'ignore'), '"'])
 
     def _handle_sighup(self, signum, frame):
         """Handle the SIGHUP signal"""
@@ -94,7 +140,7 @@ class AGI:
     def test_hangup(self):
         """This function throws AGIHangup if we have recieved a SIGHUP"""
         if self._got_sighup:
-           raise AGISIGHUPHangup("Received SIGHUP from Asterisk")
+            raise AGISIGHUPHangup("Received SIGHUP from Asterisk")
 
     def execute(self, command, *args):
         self.test_hangup()
@@ -112,7 +158,7 @@ class AGI:
     def send_command(self, command, *args):
         """Send a command to Asterisk"""
         command = command.strip()
-        command = '%s %s' % (command, ' '.join(map(str,args)))
+        command = '%s %s' % (command, ' '.join(map(str, args)))
         command = command.strip()
         if command[-1] != '\n':
             command += '\n'
@@ -120,11 +166,13 @@ class AGI:
         self.stdout.write(command)
         self.stdout.flush()
 
-    def get_result(self):
+    def get_result(self, stdin=sys.stdin):
         """Read the result of a command from Asterisk"""
         code = 0
-        result = {'result':('','')}
+        result = {'result': ('', '')}
         line = self.stdin.readline().strip()
+        if PY3:
+            if type(line) is bytes: line = line.decode('utf8')
         self.stderr.write('    RESULT_LINE: %s\n' % line)
         m = re_code.search(line)
         if m:
@@ -132,7 +180,7 @@ class AGI:
             code = int(code)
 
         if code == 200:
-            for key,value,data in re_kv.findall(response):
+            for key, value, data in re_kv.findall(response):
                 result[key] = (value, data)
 
                 # If user hangs up... we get 'hangup' in the data
@@ -149,9 +197,13 @@ class AGI:
         elif code == 520:
             usage = [line]
             line = self.stdin.readline().strip()
+            if PY3:
+                if type(line) is bytes: line = line.decode('utf8')
             while line[:3] != '520':
                 usage.append(line)
                 line = self.stdin.readline().strip()
+                if PY3:
+                    if type(line) is bytes: line = line.decode('utf8')
             usage.append(line)
             usage = '%s\n' % '\n'.join(usage)
             raise AGIUsageError(usage)
@@ -159,7 +211,7 @@ class AGI:
             raise AGIUnknownError(code, 'Unhandled code or undefined response')
 
     def _process_digit_list(self, digits):
-        if type(digits) == type([]):
+        if type(digits) is list:
             digits = ''.join(map(str, digits))
         return self._quote(digits)
 
@@ -220,7 +272,7 @@ class AGI:
     def stream_file(self, filename, escape_digits='', sample_offset=0):
         """agi.stream_file(filename, escape_digits='', sample_offset=0) --> digit
         Send the given file, allowing playback to be interrupted by the given
-        digits, if any.  escape_digits is a string '12345' or a list of
+        digits, if any.  escape_digits is a string '12345' or a list  of
         ints [1,2,3,4,5] or strings ['1','2','3'] or mixed [1,'2',3,'4']
         If sample offset is provided then the audio will seek to sample
         offset before play starts.  Returns  digit if one was pressed.
@@ -228,7 +280,8 @@ class AGI:
         extension must not be included in the filename.
         """
         escape_digits = self._process_digit_list(escape_digits)
-        response = self.execute('STREAM FILE', filename, escape_digits, sample_offset)
+        response = self.execute(
+            'STREAM FILE', filename, escape_digits, sample_offset)
         res = response['result'][0]
         if res == '0':
             return ''
@@ -267,7 +320,8 @@ class AGI:
         """
         res = self.execute('SEND IMAGE', filename)['result'][0]
         if res != '0':
-            raise AGIAppError('Channel falure on channel %s' % self.env.get('agi_channel','UNKNOWN'))
+            raise AGIAppError('Channel falure on channel %s' %
+                              self.env.get('agi_channel', 'UNKNOWN'))
 
     def say_digits(self, digits, escape_digits=''):
         """agi.say_digits(digits, escape_digits='') --> digit
@@ -328,7 +382,8 @@ class AGI:
         """
         characters = self._process_digit_list(characters)
         escape_digits = self._process_digit_list(escape_digits)
-        res = self.execute('SAY PHONETIC', characters, escape_digits)['result'][0]
+        res = self.execute(
+            'SAY PHONETIC', characters, escape_digits)['result'][0]
         if res == '0':
             return ''
         else:
@@ -374,8 +429,10 @@ class AGI:
         in seconds since the UNIX Epoch (Jan 1, 1970 00:00:00).
         """
         escape_digits = self._process_digit_list(escape_digits)
-        if format: format = self._quote(format)
-        res = self.execute('SAY DATETIME', seconds, escape_digits, format, zone)['result'][0]
+        if format:
+            format = self._quote(format)
+        res = self.execute(
+            'SAY DATETIME', seconds, escape_digits, format, zone)['result'][0]
         if res == '0':
             return ''
         else:
@@ -403,7 +460,8 @@ class AGI:
         """
         escape_digits = self._process_digit_list(escape_digits)
         if timeout:
-            response = self.execute('GET OPTION', filename, escape_digits, timeout)
+            response = self.execute(
+                'GET OPTION', filename, escape_digits, timeout)
         else:
             response = self.execute('GET OPTION', filename, escape_digits)
 
@@ -448,16 +506,20 @@ class AGI:
         self.set_extension(extension)
         self.set_priority(priority)
 
-    def record_file(self, filename, format='gsm', escape_digits='#', timeout=DEFAULT_RECORD, offset=0, beep='beep'):
-        """agi.record_file(filename, format, escape_digits, timeout=DEFAULT_TIMEOUT, offset=0, beep='beep') --> None
-        Record to a file until a given dtmf digit in the sequence is received
-        The format will specify what kind of file will be recorded.  The timeout
-        is the maximum record time in milliseconds, or -1 for no timeout. Offset
-        samples is optional, and if provided will seek to the offset without
-        exceeding the end of the file
+    def record_file(self, filename, format='gsm', escape_digits='#', timeout=DEFAULT_RECORD, offset=0, beep='beep', silence=0):
+        """agi.record_file(filename, format, escape_digits, timeout=DEFAULT_TIMEOUT, offset=0, beep='beep', silence=0) --> None
+        Record to a file until a given dtmf digit in the sequence is received. Returns
+        '-1' on hangup or error.  The format will specify what kind of file will be
+        recorded. The <timeout> is the maximum record time in milliseconds, or '-1'
+        for no <timeout>. <offset samples> is optional, and, if provided, will seek
+        to the offset without exceeding the end of the file. <silence> is the number
+        of seconds of silence allowed before the function returns despite the lack
+        of dtmf digits or reaching <timeout>. <silence> value must be preceded by
+        's=' and is also optional.
         """
         escape_digits = self._process_digit_list(escape_digits)
-        res = self.execute('RECORD FILE', self._quote(filename), format, escape_digits, timeout, offset, beep)['result'][0]
+        res = self.execute('RECORD FILE', self._quote(filename), format,
+                           escape_digits, timeout, offset, beep, ('s=%s' % silence))['result'][0]
         try:
             return chr(int(res))
         except:
@@ -512,11 +574,11 @@ class AGI:
         7 Line is busy
         """
         try:
-           result = self.execute('CHANNEL STATUS', channel)
+            result = self.execute('CHANNEL STATUS', channel)
         except AGIHangup:
-           raise
+            raise
         except AGIAppError:
-           result = {'result': ('-1','')}
+            result = {'result': ('-1', '')}
 
         return int(result['result'][0])
 
@@ -532,27 +594,28 @@ class AGI:
         the variable is not set, an empty string is returned.
         """
         try:
-           result = self.execute('GET VARIABLE', self._quote(name))
+            result = self.execute('GET VARIABLE', self._quote(name))
         except AGIResultHangup:
-           result = {'result': ('1', 'hangup')}
+            result = {'result': ('1', 'hangup')}
 
         res, value = result['result']
         return value
 
-    def get_full_variable(self, name, channel = None):
+    def get_full_variable(self, name, channel=None):
         """Get a channel variable.
 
         This function returns the value of the indicated channel variable.  If
         the variable is not set, an empty string is returned.
         """
         try:
-           if channel:
-              result = self.execute('GET FULL VARIABLE', self._quote(name), self._quote(channel))
-           else:
-              result = self.execute('GET FULL VARIABLE', self._quote(name))
+            if channel:
+                result = self.execute('GET FULL VARIABLE',
+                                      self._quote(name), self._quote(channel))
+            else:
+                result = self.execute('GET FULL VARIABLE', self._quote(name))
 
         except AGIResultHangup:
-           result = {'result': ('1', 'hangup')}
+            result = {'result': ('1', 'hangup')}
 
         res, value = result['result']
         return value
@@ -571,10 +634,14 @@ class AGI:
         is set and returns the variable in parenthesis
         example return code: 200 result=1 (testvariable)
         """
-        result = self.execute('DATABASE GET', self._quote(family), self._quote(key))
+        family = '"%s"' % family
+        key = '"%s"' % key
+        result = self.execute(
+            'DATABASE GET', self._quote(family), self._quote(key))
         res, value = result['result']
         if res == '0':
-            raise AGIDBError('Key not found in database: family=%s, key=%s' % (family, key))
+            raise AGIDBError('Key not found in database: family=%s, key=%s' %
+                             (family, key))
         elif res == '1':
             return value
         else:
@@ -585,7 +652,8 @@ class AGI:
         Adds or updates an entry in the Asterisk database for a
         given family, key, and value.
         """
-        result = self.execute('DATABASE PUT', self._quote(family), self._quote(key), self._quote(value))
+        result = self.execute('DATABASE PUT', self._quote(
+            family), self._quote(key), self._quote(value))
         res, value = result['result']
         if res == '0':
             raise AGIDBError('Unable to put vaule in databale: family=%s, key=%s, value=%s' % (family, key, value))
@@ -595,7 +663,8 @@ class AGI:
         Deletes an entry in the Asterisk database for a
         given family and key.
         """
-        result = self.execute('DATABASE DEL', self._quote(family), self._quote(key))
+        result = self.execute(
+            'DATABASE DEL', self._quote(family), self._quote(key))
         res, value = result['result']
         if res == '0':
             raise AGIDBError('Unable to delete from database: family=%s, key=%s' % (family, key))
@@ -605,7 +674,8 @@ class AGI:
         Deletes a family or specific keytree with in a family
         in the Asterisk database.
         """
-        result = self.execute('DATABASE DELTREE', self._quote(family), self._quote(key))
+        result = self.execute(
+            'DATABASE DELTREE', self._quote(family), self._quote(key))
         res, value = result['result']
         if res == '0':
             raise AGIDBError('Unable to delete tree from database: family=%s, key=%s' % (family, key))
@@ -616,7 +686,14 @@ class AGI:
         """
         self.execute('NOOP')
 
-if __name__=='__main__':
+    def exec_command(self, command, *args):
+        """Send an arbitrary asterisk command with args (even not AGI commands)"""
+        # The arguments of the command should be prepared as comma delimited, that's the way the EXEC works
+        args = ','.join(map(str, args))
+        return self.execute('EXEC', command, args)
+
+
+if __name__ == '__main__':
     agi = AGI()
     #agi.appexec('festival','Welcome to Klass Technologies.  Thank you for calling.')
     #agi.appexec('festival','This is a test of the text to speech engine.')
@@ -638,29 +715,31 @@ if __name__=='__main__':
     #agi.appexec('background','demo-congrats')
 
     try:
-        agi.appexec('backgrounder','demo-congrats')
+        agi.appexec('backgrounder', 'demo-congrats')
     except AGIAppError:
-        self.stderr.write("Handled exception for missing application backgrounder\n")
+        sys.stderr.write(
+            "Handled exception for missing application backgrounder\n")
 
-    agi.set_variable('foo','bar')
+    agi.set_variable('foo', 'bar')
     agi.get_variable('foo')
 
     try:
         agi.get_variable('foobar')
     except AGIAppError:
-        self.stderr.write("Handled exception for missing variable foobar\n")
+        sys.stderr.write("Handled exception for missing variable foobar\n")
 
     try:
         agi.database_put('foo', 'bar', 'foobar')
         agi.database_put('foo', 'baz', 'foobaz')
         agi.database_put('foo', 'bat', 'foobat')
         v = agi.database_get('foo', 'bar')
-        self.stderr.write('DBVALUE foo:bar = %s\n' % v)
+        sys.stderr.write('DBVALUE foo:bar = %s\n' % v)
         v = agi.database_get('bar', 'foo')
-        self.stderr.write('DBVALUE foo:bar = %s\n' % v)
+        sys.stderr.write('DBVALUE foo:bar = %s\n' % v)
         agi.database_del('foo', 'bar')
         agi.database_deltree('foo')
     except AGIDBError:
-        self.stderr.write("Handled exception for missing database entry bar:foo\n")
+        sys.stderr.write(
+            "Handled exception for missing database entry bar:foo\n")
 
     agi.hangup()
