@@ -1,25 +1,27 @@
 #!/usr/bin/env python
 # vim: set expandtab shiftwidth=4:
 
-from __future__ import absolute_import
-from __future__ import print_function
-from __future__ import unicode_literals
-
 """
-Python Interface for Asterisk Manager
+.. module:: manager
+   :synopsis: Python Interface for Asterisk Manager
 
 This module provides a Python API for interfacing with the asterisk manager.
 
+Example
+-------
+
+.. code-block:: python
+
    import asterisk.manager
    import sys
 
    def handle_shutdown(event, manager):
-      print ("Received shutdown event")
+      print "Recieved shutdown event"
       manager.close()
-      # we could analyze the event and reconnect here
+      # we could analize the event and reconnect here
 
    def handle_event(event, manager):
-      print ("Received event: %s" % event.name)
+      print "Recieved event: %s" % event.name
 
    manager = asterisk.manager.Manager()
    try:
@@ -36,15 +38,14 @@ This module provides a Python API for interfacing with the asterisk manager.
            response = manager.status()
 
            manager.logoff()
-       except asterisk.manager.ManagerSocketException as err:
-          errno, reason = err
-          print ("Error connecting to the manager: %s" % reason)
+       except asterisk.manager.ManagerSocketException as e:
+          print "Error connecting to the manager: %s" % e.strerror
           sys.exit(1)
-       except asterisk.manager.ManagerAuthException as reason:
-          print ("Error logging in to the manager: %s" % reason)
+       except asterisk.manager.ManagerAuthException as e:
+          print "Error logging in to the manager: %s" % e.strerror
           sys.exit(1)
-       except asterisk.manager.ManagerException as reason:
-          print ("Error: %s" % reason)
+       except asterisk.manager.ManagerException as e:
+          print "Error: %s" % e.strerror
           sys.exit(1)
 
    finally:
@@ -55,43 +56,31 @@ Remember all header, response, and event names are case sensitive.
 
 Not all manager actions are implmented as of yet, feel free to add them
 and submit patches.
+
+Specification
+-------------
 """
 
-import sys,os
+import sys
+import os
 import socket
 import threading
+from six import PY3
+from six.moves import queue
 import re
-from io import StringIO
+from types import *
 from time import sleep
-from asterisk.compat import Queue, string_types
 
 EOL = '\r\n'
 
-class _Msg(object):
-    def has_header(self, hname):
-        """Check for a header"""
-        return hname in self.headers
 
-    def get_header(self, hname, defval = None):
-        """Return the specified header"""
-        return self.headers.get(hname, defval)
-
-    def __getitem__(self, hname):
-        """Return the specified header"""
-        return self.headers[hname]
-
-    def __repr__(self):
-        return self.headers['Response']
-
-
-class ManagerMsg(_Msg):
+class ManagerMsg(object):
     """A manager interface message"""
     def __init__(self, response):
         # the raw response, straight from the horse's mouth:
         self.response = response
         self.data = ''
         self.headers = {}
-        self.multiheaders = {}
 
         # parse the response
         self.parse(response)
@@ -112,39 +101,57 @@ class ManagerMsg(_Msg):
             # 'Response', e.g., IAXpeers in Asterisk 1.4.X
             if self.has_header('ActionID'):
                 self.headers['Response'] = 'Generated Header'
-                self.multiheaders ['Response'] = ['Generated Header']
             elif '--END COMMAND--' in self.data:
                 self.headers['Event'] = 'NoClue'
-                self.multiheaders ['Event'] = ['NoClue']
             else:
                 self.headers['Response'] = 'Generated Header'
-                self.multiheaders ['Response'] = ['Generated Header']
 
     def parse(self, response):
         """Parse a manager message"""
 
         data = []
-        for n, line in enumerate (response):
+        for n, line in enumerate(response):
             # all valid header lines end in \r\n
-            if not line.endswith ('\r\n'):
+            if not line.endswith('\r\n'):
                 data.extend(response[n:])
                 break
             try:
-                k, v = (x.strip() for x in line.split(':',1))
-                if k not in self.multiheaders:
-                    self.multiheaders[k]=[]
-                self.headers[k] = v
-                self.multiheaders[k].append(v)
+                k, v = (x.strip() for x in line.split(':', 1))
+                # if header is ChanVariable it can have more that one value
+                # we store the variable in a dictionary parsed
+                if 'ChanVariable' in k:
+                    if not self.headers.has_key('ChanVariable'):
+                        self.headers['ChanVariable']={}
+                    name, value = (x.strip() for x in v.split('=',1))
+                    self.headers['ChanVariable'][name]=value
+                else:
+                    self.headers[k] = v
             except ValueError:
                 # invalid header, start of multi-line data response
                 data.extend(response[n:])
                 break
         self.data = ''.join(data)
-        if not data and 'Output' in self.multiheaders:
-            self.data = '\n'.join(self.multiheaders['Output'])
+
+    def has_header(self, hname):
+        """Check for a header"""
+        return hname in self.headers
+
+    def get_header(self, hname, defval=None):
+        """Return the specfied header"""
+        return self.headers.get(hname, defval)
+
+    def __getitem__(self, hname):
+        """Return the specfied header"""
+        return self.headers[hname]
+
+    def __repr__(self):
+        if 'Response' in self.headers:
+            return self.headers['Response']
+        else:
+            return self.headers['Event']
 
 
-class Event(_Msg):
+class Event(object):
     """Manager interface Events, __init__ expects and 'Event' message"""
     def __init__(self, message):
 
@@ -152,20 +159,33 @@ class Event(_Msg):
         self.message = message
         self.data = message.data
         self.headers = message.headers
-        self.multiheaders = message.multiheaders
 
         # if this is not an event message we have a problem
         if not message.has_header('Event'):
-            raise ManagerException('Trying to create event from non event message')
+            raise ManagerException(
+                'Trying to create event from non event message')
 
         # get the event name
         self.name = message.get_header('Event')
 
+    def has_header(self, hname):
+        """Check for a header"""
+        return hname in self.headers
+
+    def get_header(self, hname, defval=None):
+        """Return the specfied header"""
+        return self.headers.get(hname, defval)
+
+    def __getitem__(self, hname):
+        """Return the specfied header"""
+        return self.headers[hname]
+
     def __repr__(self):
         return self.headers['Event']
 
     def get_action_id(self):
-        return self.headers.get('ActionID',0000)
+        return self.headers.get('ActionID', 0000)
+
 
 class Manager(object):
     def __init__(self):
@@ -176,13 +196,11 @@ class Manager(object):
 
         # our hostname
         self.hostname = socket.gethostname()
-        # pid -- used for unique naming of ActionID
-        self.pid      = os.getpid ()
 
         # our queues
-        self._message_queue = Queue()
-        self._response_queue = Queue()
-        self._event_queue = Queue()
+        self._message_queue = queue.Queue()
+        self._response_queue = queue.Queue()
+        self._event_queue = queue.Queue()
 
         # callbacks for events
         self._event_callbacks = {}
@@ -195,12 +213,12 @@ class Manager(object):
 
         # some threads
         self.message_thread = threading.Thread(target=self.message_loop)
-        self.event_dispatch_thread = threading.Thread(target=self.event_dispatch)
+        self.event_dispatch_thread = threading.Thread(
+            target=self.event_dispatch)
 
         self.message_thread.setDaemon(True)
         self.event_dispatch_thread.setDaemon(True)
 
-
     def __del__(self):
         self.close()
 
@@ -245,31 +263,29 @@ class Manager(object):
 
         # set the action id
         if 'ActionID' not in cdict:
-            cdict['ActionID'] = '%s-%04s-%08x' % (self.hostname,
-                self.pid, self.next_seq())
+            cdict['ActionID'] = '%s-%08x' % (self.hostname, self.next_seq())
         clist = []
 
         # generate the command
         for key, value in cdict.items():
             if isinstance(value, list):
-               for item in value:
-                  item = tuple([key, item])
-                  clist.append('%s: %s' % item)
+                for item in value:
+                    item = tuple([key, item])
+                    clist.append('%s: %s' % item)
             else:
-               item = tuple([key, value])
-               clist.append('%s: %s' % item)
+                item = tuple([key, value])
+                clist.append('%s: %s' % item)
         clist.append(EOL)
         command = EOL.join(clist)
 
         # lock the socket and send our command
         try:
-            self._sock.write(command.encode('utf-8'))
+            self._sock.write(command.encode('utf8','ignore'))
             self._sock.flush()
-        except socket.error as err:
-            errno, reason = err
-            raise ManagerSocketException(errno, reason)
+        except socket.error as e:
+            raise ManagerSocketException(e.errno, e.strerror)
 
-        self._reswaiting.insert(0,1)
+        self._reswaiting.insert(0, 1)
         response = self._response_queue.get()
         self._reswaiting.pop(0)
 
@@ -284,14 +300,15 @@ class Manager(object):
         """
 
         multiline = False
+        status = False
         wait_for_marker = False
         eolcount = 0
         # loop while we are sill running and connected
         while self._running.isSet() and self._connected.isSet():
             try:
                 lines = []
-                for line in self._sock :
-                    line = line.decode('utf-8')
+                for line in self._sock:
+                    line = line.decode('utf8','ignore')
                     # check to see if this is the greeting line
                     if not self.title and '/' in line and not ':' in line:
                         # store the title of the manager we are connecting to:
@@ -299,8 +316,8 @@ class Manager(object):
                         # store the version of the manager we are connecting to:
                         self.version = line.split('/')[1].strip()
                         # fake message header
-                        lines.append ('Response: Generated Header\r\n')
-                        lines.append (line)
+                        lines.append('Response: Generated Header\r\n')
+                        lines.append(line)
                         break
                     # If the line is EOL marker we have a complete message.
                     # Some commands are broken and contain a \n\r\n
@@ -308,26 +325,36 @@ class Manager(object):
                     # have such a command where the data ends with the
                     # marker --END COMMAND--, so we ignore embedded
                     # newlines until we see that marker
-                    if line == EOL and not wait_for_marker :
+                    if line == EOL and not wait_for_marker:
                         multiline = False
                         if lines or not self._connected.isSet():
                             break
                         # ignore empty lines at start
                         continue
+                    # If the user executed the status command, it's a special
+                    # case, so we need to look for a marker.
+                    if 'status will follow' in line:
+                        status = True
+                        wait_for_marker = True
                     lines.append(line)
+
                     # line not ending in \r\n or without ':' isn't a
                     # valid header and starts multiline response
                     if not line.endswith('\r\n') or ':' not in line:
                         multiline = True
                     # Response: Follows indicates we should wait for end
                     # marker --END COMMAND--
-                    if not multiline and line.startswith('Response') and \
-                        line.split(':', 1)[1].strip() == 'Follows':
+                    if not (multiline or status) and line.startswith('Response') and \
+                            line.split(':', 1)[1].strip() == 'Follows':
                         wait_for_marker = True
                     # same when seeing end of multiline response
-                    if multiline and line.startswith('--END COMMAND--'):
+                    if multiline and (line.startswith('--END COMMAND--') or line.strip().endswith('--END COMMAND--')):
                         wait_for_marker = False
                         multiline = False
+                    # same when seeing end of status response
+                    if status and 'StatusComplete' in line:
+                        wait_for_marker = False
+                        status = False
                     if not self._connected.isSet():
                         break
                 else:
@@ -344,7 +371,6 @@ class Manager(object):
                 self._connected.clear()
                 self._message_queue.put(None)
 
-
     def register_event(self, event, function):
         """
         Register a callback for the specfied event.
@@ -373,7 +399,7 @@ class Manager(object):
         in the proper queues.
         """
 
-        # start a thread to receive data
+        # start a thread to recieve data
         t = threading.Thread(target=self._receive_data)
         t.setDaemon(True)
         t.start()
@@ -402,13 +428,11 @@ class Manager(object):
                 elif message.has_header('Response'):
                     self._response_queue.put(message)
                 else:
-                    self._response_queue.put(None)
-                    print ('No clue what we got\n%s' % message.data)
+                    print('No clue what we got\n%s' % message.data)
         finally:
             # wait for our data receiving thread to exit
             t.join()
 
-
     def event_dispatch(self):
         """This thread is responsible for dispatching events"""
 
@@ -425,12 +449,12 @@ class Manager(object):
 
             # first build a list of the functions to execute
             callbacks = (self._event_callbacks.get(ev.name, [])
-                      +  self._event_callbacks.get('*', []))
+                         + self._event_callbacks.get('*', []))
 
             # now execute the functions
             for callback in callbacks:
-               if callback(ev, self):
-                  break
+                if callback(ev, self):
+                    break
 
     def connect(self, host, port=5038):
         """Connect to the manager interface"""
@@ -439,19 +463,21 @@ class Manager(object):
             raise ManagerException('Already connected to manager')
 
         # make sure host is a string
-        assert isinstance (host, string_types)
+        assert type(host) is str
 
         port = int(port)  # make sure port is an int
 
         # create our socket and connect
         try:
             _sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-            _sock.connect((host,port))
-            self._sock = _sock.makefile(mode='rwb')
+            _sock.connect((host, port))
+            if PY3:
+                self._sock = _sock.makefile(mode='rwb', buffering=0)
+            else:
+                self._sock = _sock.makefile()
             _sock.close()
-        except socket.error as err:
-            errno, reason = err
-            raise ManagerSocketException(errno, reason)
+        except socket.error as e:
+            raise ManagerSocketException(e.errno, e.strerror)
 
         # we are connected and running
         self._connected.set()
@@ -492,26 +518,26 @@ class Manager(object):
     def login(self, username, secret):
         """Login to the manager, throws ManagerAuthException when login falis"""
 
-        cdict = {'Action':'Login'}
+        cdict = {'Action': 'Login'}
         cdict['Username'] = username
         cdict['Secret'] = secret
         response = self.send_action(cdict)
 
         if response.get_header('Response') == 'Error':
-           raise ManagerAuthException(response.get_header('Message'))
+            raise ManagerAuthException(response.get_header('Message'))
 
         return response
 
     def ping(self):
         """Send a ping action to the manager"""
-        cdict = {'Action':'Ping'}
+        cdict = {'Action': 'Ping'}
         response = self.send_action(cdict)
         return response
 
     def logoff(self):
         """Logoff from the manager"""
 
-        cdict = {'Action':'Logoff'}
+        cdict = {'Action': 'Logoff'}
         response = self.send_action(cdict)
 
         return response
@@ -519,16 +545,16 @@ class Manager(object):
     def hangup(self, channel):
         """Hangup the specified channel"""
 
-        cdict = {'Action':'Hangup'}
+        cdict = {'Action': 'Hangup'}
         cdict['Channel'] = channel
         response = self.send_action(cdict)
 
         return response
 
-    def status(self, channel = ''):
+    def status(self, channel=''):
         """Get a status message from asterisk"""
 
-        cdict = {'Action':'Status'}
+        cdict = {'Action': 'Status'}
         cdict['Channel'] = channel
         response = self.send_action(cdict)
 
@@ -537,29 +563,48 @@ class Manager(object):
     def redirect(self, channel, exten, priority='1', extra_channel='', context=''):
         """Redirect a channel"""
 
-        cdict = {'Action':'Redirect'}
+        cdict = {'Action': 'Redirect'}
         cdict['Channel'] = channel
         cdict['Exten'] = exten
         cdict['Priority'] = priority
-        if context:   cdict['Context']  = context
-        if extra_channel: cdict['ExtraChannel'] = extra_channel
+        if context:
+            cdict['Context'] = context
+        if extra_channel:
+            cdict['ExtraChannel'] = extra_channel
         response = self.send_action(cdict)
 
         return response
 
-    def originate(self, channel, exten, context='', priority='', timeout='', caller_id='', run_async=False, account='', variables={}):
+    def originate(self, channel, exten, context='', priority='', timeout='', application='', data='', caller_id='', run_async=False, earlymedia='false', account='', variables={}):
         """Originate a call"""
 
-        cdict = {'Action':'Originate'}
+        cdict = {'Action': 'Originate'}
         cdict['Channel'] = channel
         cdict['Exten'] = exten
-        if context:   cdict['Context']  = context
-        if priority:  cdict['Priority'] = priority
-        if timeout:   cdict['Timeout']  = timeout
-        if caller_id: cdict['CallerID'] = caller_id
-        if run_async: cdict['Async']    = 'yes'
-        if account:   cdict['Account']  = account
-        if variables: cdict['Variable'] = ['='.join((str(key), str(value))) for key, value in variables.items()]
+        if context:
+            cdict['Context'] = context
+        if priority:
+            cdict['Priority'] = priority
+        if timeout:
+            cdict['Timeout'] = timeout
+        if application:
+            cdict['Application'] = application
+        if data:
+            cdict['Data'] = data
+        if caller_id:
+            cdict['CallerID'] = caller_id
+        if run_async:
+            cdict['Async'] = 'yes'
+        if earlymedia:
+            cdict['EarlyMedia'] = earlymedia
+        if account:
+            cdict['Account'] = account
+        # join dict of vairables together in a string in the form of 'key=val|key=val'
+        # with the latest CVS HEAD this is no longer necessary
+        # if variables: cdict['Variable'] = '|'.join(['='.join((str(key), str(value))) for key, value in variables.items()])
+        if variables:
+            cdict['Variable'] = ['='.join(
+                (str(key), str(value))) for key, value in variables.items()]
 
         response = self.send_action(cdict)
 
@@ -568,7 +613,7 @@ class Manager(object):
     def mailbox_status(self, mailbox):
         """Get the status of the specfied mailbox"""
 
-        cdict = {'Action':'MailboxStatus'}
+        cdict = {'Action': 'MailboxStatus'}
         cdict['Mailbox'] = mailbox
         response = self.send_action(cdict)
 
@@ -577,7 +622,7 @@ class Manager(object):
     def command(self, command):
         """Execute a command"""
 
-        cdict = {'Action':'Command'}
+        cdict = {'Action': 'Command'}
         cdict['Command'] = command
         response = self.send_action(cdict)
 
@@ -586,16 +631,17 @@ class Manager(object):
     def extension_state(self, exten, context):
         """Get the state of an extension"""
 
-        cdict = {'Action':'ExtensionState'}
+        cdict = {'Action': 'ExtensionState'}
         cdict['Exten'] = exten
         cdict['Context'] = context
         response = self.send_action(cdict)
 
         return response
 
-    def playdtmf (self, channel, digit) :
+    def playdtmf(self, channel, digit):
         """Plays a dtmf digit on the specified channel"""
-        cdict = {'Action':'PlayDTMF'}
+
+        cdict = {'Action': 'PlayDTMF'}
         cdict['Channel'] = channel
         cdict['Digit'] = digit
         response = self.send_action(cdict)
@@ -605,33 +651,83 @@ class Manager(object):
     def absolute_timeout(self, channel, timeout):
         """Set an absolute timeout on a channel"""
 
-        cdict = {'Action':'AbsoluteTimeout'}
+        cdict = {'Action': 'AbsoluteTimeout'}
         cdict['Channel'] = channel
         cdict['Timeout'] = timeout
         response = self.send_action(cdict)
-
         return response
 
     def mailbox_count(self, mailbox):
-        cdict = {'Action':'MailboxCount'}
+        cdict = {'Action': 'MailboxCount'}
         cdict['Mailbox'] = mailbox
         response = self.send_action(cdict)
-
         return response
 
     def sippeers(self):
-        cdict = {'Action' : 'Sippeers'}
+        cdict = {'Action': 'Sippeers'}
         response = self.send_action(cdict)
         return response
 
     def sipshowpeer(self, peer):
-        cdict = {'Action' : 'SIPshowpeer'}
+        cdict = {'Action': 'SIPshowpeer'}
         cdict['Peer'] = peer
         response = self.send_action(cdict)
         return response
 
+    def sipshowregistry(self):
+        cdict = {'Action': 'SIPShowregistry'}
+        response = self.send_action(cdict)
+        return response
 
-class ManagerException(Exception): pass
-class ManagerSocketException(ManagerException): pass
-class ManagerAuthException(ManagerException): pass
+    def iaxregistry(self):
+        cdict = {'Action': 'IAXregistry'}
+        response = self.send_action(cdict)
+        return response
 
+    def reload(self, module):
+        """ Reloads config for a given module """
+
+        cdict = {'Action': 'Reload'}
+        cdict['Module'] = module
+        response = self.send_action(cdict)
+        return response
+
+    def dbdel(self, family, key):
+        cdict = {'Action': 'DBDel'}
+        cdict['Family'] = family
+        cdict['Key'] = key
+        response = self.send_action(cdict)
+        return response
+
+    def dbdeltree(self, family, key):
+        cdict = {'Action': 'DBDelTree'}
+        cdict['Family'] = family
+        cdict['Key'] = key
+        response = self.send_action(cdict)
+        return response
+
+    def dbget(self, family, key):
+        cdict = {'Action': 'DBGet'}
+        cdict['Family'] = family
+        cdict['Key'] = key
+        response = self.send_action(cdict)
+        return response
+
+    def dbput(self, family, key, val):
+        cdict = {'Action': 'DBPut'}
+        cdict['Family'] = family
+        cdict['Key'] = key
+        cdict['Val'] = val
+        response = self.send_action(cdict)
+        return response
+
+class ManagerException(Exception):
+    pass
+
+
+class ManagerSocketException(ManagerException):
+    pass
+
+
+class ManagerAuthException(ManagerException):
+    pass
